from nlp_tools.layers.non_masking_layer import NonMaskingLayer
from nlp_tools.layers.conditional_random_field import KConditionalRandomField
from nlp_tools.layers.kmax_pool_layer import KMaxPoolingLayer
from nlp_tools.layers.att_wgt_avg_layer import AttentionWeightedAverageLayer
from nlp_tools.layers.behdanau_attention import BahdanauAttention
from nlp_tools.loss.r_drop_loss import RDropLoss

from nlp_tools.loss import TextGenerateCrossEntropy
from typing import Dict, Any

from tensorflow import keras
L = keras.layers
L.BahdanauAttention = BahdanauAttention
L.KConditionalRandomField = KConditionalRandomField


L.TextGenerateCrossEntropy = TextGenerateCrossEntropy



from transformers import TFBertModel
def resigter_custom_layers(custom_objects: Dict[str, Any]) -> Dict[str, Any]:
    custom_objects['KConditionalRandomField'] = KConditionalRandomField
    custom_objects['BahdanauAttention'] = BahdanauAttention
    custom_objects['loss'] = KConditionalRandomField.loss

    custom_objects['TextGenerateCrossEntropy'] = TextGenerateCrossEntropy
    custom_objects['r_drop_loss'] = RDropLoss(loss_func=None)

    custom_objects['TFBertModel'] = TFBertModel


    return custom_objects


from typing import Tuple, List
import random



class LCQMC(object):
    __corpus_name__ = 'lcqmc'

    @classmethod
    def load_data(cls,
                  file_path: str ,
                  shuffle: bool = True):
        """
                Load dataset as sequence labeling format, char level tokenized

                features: ``[['海', '钓', '比', '赛', '地', '点', '在', '厦', '门', ...], ...]``

                labels: ``[[海', '钓', '比', '赛', '地', '点', , ...], ...]``

                Args:
                    file_path: file ablsute path
                    shuffle: should shuffle or not, default True.

                Returns:
                    dataset_features and dataset labels
                """
        data = []
        with open(file_path,'r',encoding='utf-8') as fread:
            for index, line in enumerate(fread):
                x1,x2,label = line.strip().split("\t")
                data.append(([x1,x2],str(label)))

        if shuffle:
            random.shuffle(data)
        return data




from abc import ABC
from typing import Iterable, Iterator, TYPE_CHECKING
from typing import List, Any, Tuple

import numpy as np
import tensorflow as tf




class DataGenerator(object):
    """数据生成器模版
    """
    def __init__(self, data, batch_size=32, buffer_size=None):
        self.data = data
        self.batch_size = batch_size
        if hasattr(self.data, '__len__'):
            self.steps = len(self.data) // self.batch_size
            if len(self.data) % self.batch_size != 0:
                self.steps += 1
        else:
            self.steps = None
        self.buffer_size = buffer_size or batch_size * 1000

    def __len__(self):
        return self.steps

    def sample(self, random=False):
        """采样函数，每个样本同时返回一个is_end标记
        """
        if random:
            if self.steps is None:

                def generator():
                    caches, isfull = [], False
                    for d in self.data:
                        caches.append(d)
                        if isfull:
                            i = np.random.randint(len(caches))
                            yield caches.pop(i)
                        elif len(caches) == self.buffer_size:
                            isfull = True
                    while caches:
                        i = np.random.randint(len(caches))
                        yield caches.pop(i)

            else:

                def generator():
                    for i in np.random.permutation(len(self.data)):
                        yield self.data[i]

            data = generator()
        else:
            data = iter(self.data)

        d_current = next(data)
        for d_next in data:
            yield d_current
            d_current = d_next

        yield d_current

    def __iter__(self, random=False):
        raise NotImplementedError

    def forfit(self, random=False):
        while True:
            for d in self.__iter__(random):
                yield d

    def to_dataset(self, types, shapes, names=None, padded_batch=False):
        """转为tf.data.Dataset格式
        如果传入names的话，自动把数据包装成dict形式。
        """
        if names is None:

            generator = self.forfit

        else:

            if type(names) == str:
                warps = lambda k, v: {k: v}
            elif type(names[0]) == str:
                warps = lambda k, v: dict(zip(k, v))
            else:
                warps = lambda k, v: tuple(
                    dict(zip(i, j)) for i, j in zip(k, v)
                )

            def generator():
                for d in self.forfit():
                    yield warps(names, d)

            types = warps(names, types)
            shapes = warps(names, shapes)

        if padded_batch:
            dataset = tf.data.Dataset.from_generator(
                generator, output_types=types
            )
            dataset = dataset.padded_batch(self.batch_size, shapes)
        else:
            dataset = tf.data.Dataset.from_generator(
                generator, output_types=types, output_shapes=shapes
            )
            dataset = dataset.batch(self.batch_size)

        return dataset


class ABCGenerator(Iterable, ABC):
    def __init__(self, buffer_size: int = 2000) -> None:
        self.buffer_size = buffer_size

    def __iter__(self) -> Iterator[Tuple[Any, Any]]:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def sample(self) -> Iterator[Tuple[Any, Any]]:
        buffer, is_full = [], False
        for sample in self:
            buffer.append(sample)
            if is_full:
                i = np.random.randint(len(buffer))
                yield buffer.pop(i)
            elif len(buffer) == self.buffer_size:
                is_full = True
        while buffer:
            i = np.random.randint(len(buffer))
            yield buffer.pop(i)


class CorpusGenerator(ABCGenerator):

    def __init__(self,
                 x_data: List,
                 y_data: List,
                 *,
                 buffer_size: int = 2000) -> None:
        super(CorpusGenerator, self).__init__(buffer_size=buffer_size)
        self.x_data = x_data
        self.y_data = y_data
        self.buffer_size = buffer_size

    def __iter__(self) -> Iterator[Tuple[Any, Any]]:
        for i in range(len(self.x_data)):
            yield self.x_data[i], self.y_data[i]

    def __len__(self) -> int:
        return len(self.x_data)




class BatchGenerator(DataGenerator):
    """数据生成器
    """
    def __init__(self,data,text_processor,label_processor,seq_length=None,batch_size=64, buffer_size=None,use_rdrop=True):
        super(BatchGenerator,self).__init__(data,batch_size,buffer_size)
        self.text_processor = text_processor
        self.label_processor = label_processor
        self.seq_length = seq_length
        self.use_rdrop = use_rdrop


    def __iter__(self, random=False):
        batch_x, batch_y = [], []
        for (x, y) in self.sample():
            if self.use_rdrop:
                batch_size = self.batch_size * 2
                for i in range(2):
                    batch_x.append(x)
                    batch_y.append(y)
            else:
                batch_x.append(x)
                batch_y.append(y)
                batch_size = self.batch_size



            if len(batch_x) == batch_size:
                x_tensor = self.text_processor.transform(batch_x,seq_length=self.seq_length)
                # x_tensor [x_token, segment]

                max_sequence_len = x_tensor[0].shape[1]
                if self.seq_length:
                    max_sequence_len = self.seq_length

                y_tensor = self.label_processor.transform(batch_y,
                                                          seq_length=max_sequence_len)
                yield x_tensor, y_tensor
                batch_x, batch_y = [], []

        if batch_x:
            x_tensor = self.text_processor.transform(batch_x,
                                                     seq_length=self.seq_length,
                                                     )
            # x_tensor [x_token, segment]
            max_sequence_len = x_tensor[0].shape[1]
            if self.seq_length:
                max_sequence_len = self.seq_length

            y_tensor = self.label_processor.transform(batch_y,
                                                      seq_length=max_sequence_len)
            yield x_tensor, y_tensor




class GlobalPointGenerator(DataGenerator):
    """数据生成器
    """
    def __init__(self,data,text_processor,label_processor,seq_length=None,batch_size=64, buffer_size=None,use_rdrop=True):
        super(GlobalPointGenerator,self).__init__(data,batch_size,buffer_size)
        self.text_processor = text_processor
        self.label_processor = label_processor
        self.seq_length = seq_length

    def __iter__(self, random=False):
        batch_x, batch_y = [], []
        for (x, y) in self.sample():
            batch_x.append(x)
            batch_y.append(y)
            if len(batch_x) == self.batch_size :
                x_tensor = self.text_processor.transform(batch_x,
                                                         seq_length=self.seq_length,
                                                         )
                # x_tensor [x_token, segment]
                max_sequence_len = x_tensor[0].shape[1]
                if self.seq_length:
                    max_sequence_len = self.seq_length

                y_tensor = self.label_processor.transform(batch_y,
                                                          seq_length=max_sequence_len)
                yield x_tensor, y_tensor
                batch_x, batch_y = [], []
        if batch_x:
            x_tensor = self.text_processor.transform(batch_x,
                                                     seq_length=self.seq_length,
                                                     )
            # x_tensor [x_token, segment]
            max_sequence_len = x_tensor[0].shape[1]
            if self.seq_length:
                max_sequence_len = self.seq_length

            y_tensor = self.label_processor.transform(batch_y,
                                                      seq_length=max_sequence_len)
            yield x_tensor, y_tensor



class Seq2SeqDataSet(Iterable):
    def __init__(self,
                 corpus: CorpusGenerator,
                 *,
                 batch_size: int = 64,
                 encoder_processor,
                 decoder_processor,
                 encoder_seq_length: int = None,
                 decoder_seq_length: int = None,
                 encoder_segment: bool = False,
                 decoder_segment: bool = False):
        self.corpus = corpus

        self.encoder_processor = encoder_processor
        self.decoder_processor = decoder_processor

        self.encoder_seq_length = encoder_seq_length
        self.decoder_seq_length = decoder_seq_length

        self.encoder_segment = encoder_segment
        self.decoder_segment = decoder_segment

        self.batch_size = batch_size

    def __len__(self) -> int:
        return max(len(self.corpus) // self.batch_size, 1)

    def __iter__(self) -> Iterator:
        batch_x, batch_y = [], []
        for x, y in self.corpus.sample():
            batch_x.append(x)
            batch_y.append(y)
            if len(batch_x) == self.batch_size:
                x_tensor = self.encoder_processor.transform(batch_x,
                                                            seq_length=self.encoder_seq_length,
                                                            segment=self.encoder_segment)
                y_tensor = self.decoder_processor.transform(batch_y,
                                                            seq_length=self.decoder_seq_length,
                                                            segment=self.encoder_segment)
                yield x_tensor, y_tensor
                batch_x, batch_y = [], []

    def take(self, batch_count: int = None) -> Any:
        x_shape = [self.batch_size, self.encoder_seq_length]
        y_shape = [self.batch_size, self.decoder_seq_length]
        dataset = tf.data.Dataset.from_generator(self.__iter__,
                                                 output_types=(tf.int64, tf.int64),
                                                 output_shapes=(x_shape, y_shape))
        dataset = dataset.repeat()
        dataset = dataset.prefetch(50)
        if batch_count is None:
            batch_count = len(self)
        return dataset.take(batch_count)





class BertGeneratorDataSet(Iterable):
    '''
    BERT generator data
    '''
    def __init__(self,
                 corpus: CorpusGenerator,
                 *,
                 batch_size: int = 64,
                 seq_length: int = None,
                 max_position: int = None,
                 segment: bool = True,
                 text_processor: 'ABCProcessor'):
        self.corpus = corpus

        self.text_processor = text_processor
        self.seq_length = seq_length
        self.max_position = max_position
        self.segment = segment
        self.batch_size = batch_size

    def __len__(self) -> int:
        return max(len(self.corpus) // self.batch_size, 1)

    def __iter__(self) -> Iterator:
        batch_x = []
        for x, y in self.corpus.sample():
            if y:
                batch_x.append((x,y))
            else:
                batch_x.append(x)
            if len(batch_x) == self.batch_size:
                x_tensor = self.text_processor.transform(batch_x,
                                                            seq_length=self.seq_length,
                                                            segment=self.segment)

                yield x_tensor,None
                batch_x = []

    def take(self, batch_count: int = None) -> Any:
        # x_shape = [2,self.batch_size, self.seq_length]
        # dataset = tf.data.Dataset.from_generator(self.__iter__,
        #                                          output_types=(tf.int64)
        #                                          )
        # dataset = dataset.repeat()
        # dataset = dataset.prefetch(50)
        # if batch_count is None:
        #     batch_count = len(self)
        # return dataset.take(batch_count)
        while True:
            for d in self.__iter__():
                yield d





from typing import Dict,Any
from tensorflow import keras
from nlp_tools.layers import L
from nlp_tools.tasks.simalary.abc_model import ABCSimilaryModel

class SimpleBertModel(ABCSimilaryModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'dropout':{'rate':0.1},
            'layer_output':{
                'activation':'softmax'
            }
        }

    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size
        config = self.hyper_parameters


        layer_stack = [
            L.Dropout(**config['dropout']),
            L.Dense(output_dim, **config['layer_output']),
        ]

        tensor = self.embedding.embed_model.output

        for layer in layer_stack:
            tensor = layer(tensor)

        self.tf_model = keras.Model(self.embedding.embed_model.inputs,tensor)





from typing import Dict,Any

from nlp_tools.tokenizer.base_tokenizer import ABCTokenizer
from bert4keras.tokenizers import load_vocab,Tokenizer






class BertTokenizer(Tokenizer,ABCTokenizer):

    """
    Bert Like Tokenizer, ref: https://github.com/CyberZHG/keras-bert/blob/master/keras_bert/tokenizer.py

    """
    def to_dict(self) -> Dict[str, Any]:
        data = super(BertTokenizer, self).to_dict()
        data['config']['token_dict'] = self._token_dict
        data['config']['do_lower_case'] = self._do_lower_case
        data['config']['simplified'] = self.simplified
        data['config']['keep_tokens'] = self.keep_tokens
        data['config']['max_position'] = self.max_position
        return data

    def __init__(self,token_dict=None, do_lower_case=False,simplified=False,keep_tokens = None,max_position = 512):
        self.simplified = simplified
        self.keep_tokens = keep_tokens
        self.max_position = max_position
        self.token_dict = token_dict
        self.do_lower_case = do_lower_case
        if type(token_dict) != dict:
            if simplified:
                self.token_dict, self.keep_tokens = load_vocab(
                    dict_path=token_dict,
                    simplified=simplified,
                    startswith=['[PAD]', '[UNK]', '[CLS]', '[SEP]'],
                )
            else:
                self.token_dict = load_vocab(
                    dict_path=token_dict,
                )
        super(BertTokenizer,self).__init__(token_dict=self.token_dict,do_lower_case=self.do_lower_case)


    def tokenize(self, text, maxlen=None,**kwargs):
        if not maxlen:
            maxlen = self.max_position
        elif maxlen > self.max_position:
            maxlen = self.max_position

        if type(text) == str:
            text = super(BertTokenizer, self).tokenize(text,maxlen=maxlen)
        elif type(text) == list:
            if len(text) > maxlen -2:
                text = text[:maxlen-2]

            # 如果不以cls开头，则需要补充cls
            if text[0] != '[CLS]':
                text = ['[CLS]'] + text + ['[SEP]']
        return text

    def encode(self,
                text,
                second_text=None,
                maxlen=None,
                pattern='S*E*E',
                truncate_from='right'):
        assert type(text) in [list, str]
        if maxlen and maxlen > self.max_position:
            maxlen = self.max_position

        text = self.tokenize(text)
        tokens, segment_ids = super(BertTokenizer, self).encode(text, maxlen=maxlen,second_text=second_text,pattern=pattern,truncate_from=truncate_from)
        return tokens, segment_ids

    def id_to_token(self, id):
        return super(BertTokenizer, self).id_to_token(id)

if __name__ == '__main__':
    import os
    bert_model_path = r'/home/qiufengfeng/nlp/pre_trained_model/chinese_L-12_H-768_A-12/chinese_L-12_H-768_A-12/'
    bert_model_token_path = os.path.join(bert_model_path, 'vocab.txt')
    bert_tokenizer = BertTokenizer(bert_model_token_path)
    print(bert_tokenizer.tokenize("[CLS]"))


import collections
import operator
from typing import Dict,List,Any,Optional,Union

import numpy as np
import tqdm
from tensorflow.keras.preprocessing.sequence import pad_sequences

from nlp_tools.generators import CorpusGenerator
from nlp_tools.logger import  logger
from nlp_tools.processors.abc_processor import ABCProcessor
from nlp_tools.types import TextSamplesVar


class SimilarySequenceProcessor(ABCProcessor):
    """
        Generic processors for the sequence samples.
        """

    def to_dict(self) -> Dict[str,Any]:
        data = super(SimilarySequenceProcessor,self).to_dict()
        data['config'].update({
            'build_in_vocab':self.build_in_vocab,
            'min_count':self.min_count,
            'conbime_sentence':False
        })
        return data

    def __init__(self,
                 build_in_vocab: str = 'text',
                 min_count: int = 3,
                 conbime_sentence:bool = False,
                 **kwargs: Any) -> None:
        """
                Args:
                    vocab_dict_type: initial vocab dict type, one of `text` `labeling`.
                    **kwargs:
                """
        super(SimilarySequenceProcessor,self).__init__(**kwargs)

        self.build_in_vocab = build_in_vocab
        self.min_count = min_count
        self.conbime_sentence = conbime_sentence

        if build_in_vocab == 'text':
            self._initial_vocab_dict = {
                self.token_pad: 0,
                self.token_unk: 1,
                self.token_bos: 2,
                self.token_eos: 3
            }
        else:
            self._initial_vocab_dict = {}

        self._showed_seq_len_warning = False

    def build_vocab_generator(self,generators: List[CorpusGenerator]) -> None:
        if not self.vocab2idx:
            vocab2idx = self._initial_vocab_dict

            token2count: Dict[str,int] = {}

            for gen in generators:
                for sentence,label in tqdm.tqdm(gen,desc="Preparing text vocab dict"):
                    target_1,target_2 = sentence
                    for token in self.tokenizer.tokenize(target_1):
                        count = token2count.get(token,0)
                        token2count[token] = count+ 1

                    for token in self.tokenizer.tokenize(target_2):
                        count = token2count.get(token,0)
                        token2count[token] = count+ 1



            sorted_token2count = sorted(token2count.items(),
                                        key=operator.itemgetter(1),
                                        reverse=True)
            token2count = collections.OrderedDict(sorted_token2count)

            for token,token_count in token2count.items():
                if token not in vocab2idx and token_count >= self.min_count:
                    vocab2idx[token] = len(vocab2idx)
            self.vocab2idx = vocab2idx
            self.idx2vocab = dict([(v,k) for k,v in self.vocab2idx.items()])

            top_k_vocab = [k for (k,v) in list(self.vocab2idx.items())[:10]]
            logger.debug(f"--- Build vocab dict finished, Total: {len(self.vocab2idx)} ---")
            logger.debug(f"Top-10: {top_k_vocab}")

    def transform(self,
                  samples: TextSamplesVar,
                  *,
                  seq_length: int = None,
                  max_position: int = None,
                  segment: bool = False ) -> np.ndarray:
        tokenized_sentences = []
        for pair in samples:
            seq_1,seq_2 = pair
            tokenized_sentences.append([self.tokenizer.tokenize(seq_1),self.tokenizer.tokenize(seq_2)])


        seq_length_from = ""
        if seq_length is None:
            seq_length_from = 'max length of the samples'
            seq_1_list = [i[0] for i in tokenized_sentences]
            seq_2_list = [i[1] for i in tokenized_sentences]

            max_1 = max([len(i) for i in seq_1_list])
            max_2 = max([len(i) for i in seq_2_list])
            if self.conbime_sentence:
                seq_length = max_1 + max_2
            else:
                seq_length = max(max_1,max_2)
        if max_position is not None and max_position < seq_length:
            seq_length_from = 'max embedding seq length'
            seq_length = max_position

        if seq_length_from and not self._showed_seq_len_warning:
            logger.warning(
                f'Sequence length is None, will use the {seq_length_from}, which is {seq_length}')
            self._showed_seq_len_warning = True

        numerized_samples = []
        segment_ids = []

        if self.conbime_sentence:
            for seq in samples:
                seq_tokens,seq_segment_ids = self.tokenizer.encode(seq[0],seq[1],maxlen=seq_length)

                numerized_samples.append(seq_tokens)
                segment_ids.append(seq_segment_ids)

            token_ids = pad_sequences(numerized_samples, seq_length, padding='post', truncating='post')
            segment_ids = pad_sequences(segment_ids, seq_length, padding='post', truncating='post')
        else:
            seq_list_1 = []
            seq_list_2 = []
            for seq in samples:
                if hasattr(self.tokenizer,"encode"):
                    seq_tokens_1, seq_segment_ids_1 = self.tokenizer.encode(seq[0],maxlen=seq_length)
                    seq_tokens_2, seq_segment_ids_2 = self.tokenizer.encode(seq[1], maxlen=seq_length)
                else:
                    seq_tokens_1,seq_tokens_2 = seq
                    seq_tokens_1, seq_tokens_2 = self.tokenizer.tokenize(seq_tokens_1),self.tokenizer.tokenize(seq_tokens_2)

                    if self.token_bos in self.vocab2idx:
                        seq_tokens_1 = [self.token_bos] + seq_tokens_1 + [self.token_eos]
                        seq_tokens_2 = [self.token_bos] + seq_tokens_2 + [self.token_eos]
                    else:
                        seq_tokens_1 = [self.token_pad] + seq_tokens_1 + [self.token_pad]
                        seq_tokens_2 = [self.token_pad] + seq_tokens_2 + [self.token_pad]

                    if self.token_unk in self.vocab2idx:
                        unk_index = self.vocab2idx[self.token_unk]
                        seq_tokens_1 = [self.vocab2idx.get(token, unk_index) for token in seq_tokens_1]
                        seq_tokens_2 = [self.vocab2idx.get(token, unk_index) for token in seq_tokens_2]
                    else:
                        seq_tokens_1 = [self.vocab2idx[token] for token in seq_tokens_1]
                        seq_tokens_2 = [self.vocab2idx[token] for token in seq_tokens_2]
                seq_list_1.append(seq_tokens_1)
                seq_list_2.append(seq_tokens_2)


            seq_list_1 = pad_sequences(seq_list_1, seq_length, padding='post', truncating='post')
            seq_segment_ids_1  = np.zeros(seq_list_1.shape,dtype=np.int32)
            seq_list_2 = pad_sequences(seq_list_2, seq_length, padding='post', truncating='post')
            seq_segment_ids_2  = np.zeros(seq_list_2.shape,dtype=np.int32)


            #numerized_samples.append([seq_list_1,seq_list_2])
            segment_ids.append([seq_segment_ids_1,seq_segment_ids_2])
            token_ids = [seq_list_1,seq_list_2]
            segment_ids = [seq_segment_ids_1,seq_segment_ids_2]



        if segment:
            return token_ids,segment_ids
        else:
            return token_ids

    def inverse_transform(self,
                          labels: Union[List[List[int]], np.ndarray],
                          *,
                          lengths: List[int] = None,
                          threshold: float = 0.5,
                          **kwargs: Any) -> List[List[str]]:
        result = []
        for index,seq in enumerate(labels):
            labels_ = []
            for idx in seq:
                labels_.append(self.idx2vocab[idx])

            if lengths is not None:
                labels_ = labels_[1:lengths[index] + 1]
            else:
                labels_ = labels_[1:-1]
            result.append(labels_)
        return result


from tensorflow.keras.layers import Layer,Dense
import tensorflow.keras.backend as K
import tensorflow as tf

from nlp_tools.utils.keras_utils import sequence_masking

class SinusoidalPositionEmbedding(Layer):
    """定义Sin-Cos位置Embedding
        """
    def __init__(self
                 ,output_dim,merge_mode='add',custom_position_ids = False,**kwargs):
        super(SinusoidalPositionEmbedding,self).__init__(**kwargs)
        self.output_dim = output_dim
        self.merge_mode = merge_mode
        self.custom_position_ids = custom_position_ids

    def call(self,inputs,**kwargs):
        """如果custom_position_ids，那么第二个输入为自定义的位置id
                """
        if self.custom_position_ids:
            seq_len = K.shape(inputs)[1]
            inputs,position_ids = inputs
            if 'float' not in K.dtype(position_ids):
                position_ids = K.cast(position_ids, K.floatx())
        else:
            input_shape = K.shape(inputs)
            batch_size, seq_len = input_shape[0], input_shape[1]
            position_ids = K.arange(0,seq_len,dtype=K.floatx())[None]

        indices = K.arange(0,self.output_dim // 2, dtype=K.floatx())
        indices = K.pow(10000.0, -2 * indices / self.output_dim)
        embeddings = tf.einsum('bn,d->bnd',position_ids,indices)
        embeddings = K.stack([K.sin(embeddings),K.cos(embeddings)],axis=-1)
        embeddings = K.reshape(embeddings,(-1,seq_len,self.output_dim))

        if self.merge_mode == 'add':
            return inputs + embeddings
        elif self.merge_mode == 'mul':
            return inputs * (embeddings + 1.0)
        elif self.merge_mode == 'zero':
            return embeddings
        else:
            if not self.custom_position_ids:
                embeddings = K.tile(embeddings,[batch_size,1,1])
            return K.concatenate([inputs,embeddings])

    def compute_output_shape(self, input_shape):
        if self.custom_position_ids:
            input_shape = input_shape[0]

        if self.merge_mode in ['add','mul','zero']:
            return input_shape[:2] + (self.output_dim,)
        else:
            return input_shape[:2] + (input_shape[2] + self.output_dim,)

    def get_config(self):
        config = {
            'output_dim': self.output_dim,
            'merge_mode': self.merge_mode,
            'custom_position_ids': self.custom_position_ids
        }
        base_config = super(SinusoidalPositionEmbedding, self).get_config()
        return  dict(list(base_config.items()) + list(config.items()))



class GloablPointerLayer(Layer):
    """全局指针模块
        将序列的每个(start, end)作为整体来进行判断
        """
    def __init__(self,heads, head_size,RoPE =True,**kwargs):
        super(GloablPointerLayer,self).__init__(**kwargs)
        self.heads = heads
        self.head_size = head_size
        self.RoPE = RoPE

    def build(self,input_shape):
        super(GloablPointerLayer,self).build(input_shape)
        self.dense = Dense(self.head_size * self.heads * 2)

    def compute_mask(self, inputs, mask=None):
        return None

    def call(self, inputs, mask=None):
        inputs = self.dense(inputs)
        inputs = tf.split(inputs,self.heads, axis=-1)
        inputs = K.stack(inputs,axis=-2)
        qw,kw = inputs[...,:self.head_size],inputs[...,self.head_size:]

        # RoPE编码
        if self.RoPE:
            pos = SinusoidalPositionEmbedding(self.head_size,'zero')(inputs)
            cos_pos = K.repeat_elements(pos[...,None,1::2],2,-1)
            sin_pos = K.repeat_elements(pos[...,None,::2],2,-1)
            qw2 = K.stack([-qw[...,1::2],qw[...,::2]],4)
            qw2 = K.reshape(qw2,K.shape(qw))
            qw = qw * cos_pos + qw2 * sin_pos

            kw2 = K.stack([-kw[...,1::2],kw[...,::2]],4)
            kw2 = K.reshape(kw2,K.shape(kw))
            kw  = kw * cos_pos + kw2 * sin_pos

        # 计算内积
        logits = tf.einsum('bmhd,bnhd->bhmn',qw,kw)

        # 排除padding
        logits = sequence_masking(logits,mask,'-inf',2)
        logits = sequence_masking(logits,mask,'-inf',3)

        # 排除下三角
        mask = tf.linalg.band_part(K.ones_like(logits),0,-1)
        logits = logits - ( 1 - mask) * 1e12

        # scale返回
        return logits / self.head_size ** 0.5

    def compute_output_shape(self, input_shape):
        return (input_shape[0] , self.heads,input_shape[1] ,input_shape[1])

    def get_config(self):
        config = {
            'heads': self.heads,
            'head_size': self.head_size,
            'RoPE':self.RoPE
        }
        base_config = super(GloablPointerLayer,self).get_config()
        return dict(list(base_config.items()) + list(config.items()))





if __name__ == '__main__':
    import numpy as np
    a = np.arange(0,10).reshape((2,1,1,5))
    print(np.repeat(a,2,-1).shape)



import nlp_tools
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K

L = keras.layers
InputSpec = L.InputSpec

class KMaxPoolingLayer(L.Layer):
    '''
        K-max pooling layer that extracts the k-highest activation from a sequence (2nd dimension).
        TensorFlow backend.
        # Arguments
            k: An int scale,
                indicate k max steps of features to pool.
            sorted: A bool,
                if output is sorted (default) or not.
            data_format: A string,
                one of `channels_last` (default) or `channels_first`.
                The ordering of the dimensions in the inputs.
                `channels_last` corresponds to inputs with shape
                `(batch, steps, features)` while `channels_first`
                corresponds to inputs with shape
                `(batch, features, steps)`.
        # Input shape
            - If `data_format='channels_last'`:
                3D tensor with shape:
                `(batch_size, steps, features)`
            - If `data_format='channels_first'`:
                3D tensor with shape:
                `(batch_size, features, steps)`
        # Output shape
            3D tensor with shape:
            `(batch_size, top-k-steps, features)`
        '''
    def __init__(self,k=1,sorted=True,data_format='channels_last',**kwargs):
        super(KMaxPoolingLayer,self).__init__(**kwargs)
        self.input_spec = InputSpec(ndim=3)
        self.k = k
        self.sorted = sorted
        if data_format.lower() in ['channels_first','channels_last']:
            self.data_format = data_format.lower()
        else:
            self.data_format = K.image_data_format()

    def compute_output_shape(self, input_shape):
        if self.data_format == 'channels_first':
            return (input_shape[0],self.k,input_shape[1])
        else:
            return (input_shape[0],self.k,input_shape[2])

    def call(self,inputs,**kwargs):
        if self.data_format == 'channels_last':
            # swap las
            # text_generator two dimensions since top_k will be applied along the last dimension
            shifted_input = tf.transpose(inputs, [0, 2, 1])

            # extract top_k, returns two tensors [values, indices]
            top_k = tf.nn.top_k(shifted_input,k = self.k,sorted=self.sorted)[0]
        else:
            top_k = tf.nn.top_k(inputs,k=self.k,sorted=self.sorted)[0]

        return tf.transpose(top_k,[0,2,1])

    def get_config(self):
        config = {
            'k':self.k,
            'sorted':self.sorted,
            'data_format':self.data_format
        }
        base_config = super(KMaxPoolingLayer,self).get_config()
        return dict(list(base_config.items()) + list(config.items))


KMaxPooling = KMaxPoolingLayer
KMaxPoolLayer = KMaxPoolingLayer

nlp_tools.custom_objects['KMaxPoolingLayer'] = KMaxPoolingLayer
nlp_tools.custom_objects['KMaxPooling'] = KMaxPooling
nlp_tools.custom_objects['KMaxPoolLayer'] = KMaxPoolLayer







from typing import Dict,Any
from tensorflow import keras
from nlp_tools.layers import L
from nlp_tools.tasks.simalary.abc_model import ABCSimilaryModel
import tensorflow as tf
from nlp_tools.layers.similary import MatchingLayer







class ArciiModel(ABCSimilaryModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            "conv_1d_layer":{
                'filters':64,
                'kernel_size': 3,
                'padding':'same',
            },
            'matching_layer':{
                "normalize":True,
                'matching_type':'dot'
            },
            'num_blocks':3,
            'conv_block_1':{
                'filters': 32,
                'kernel_size': 3,
                'strides': 1,
                'padding': 'same',
                'activation': 'relu',
            },
            'pool_block_1':{
                "pool_size":2
            },
            'conv_block_2':{
                'filters': 64,
                'kernel_size': 3,
                'strides': 1,
                'padding': 'same',
                'activation': 'relu',
            },
            'pool_block_2': {
                "pool_size": 2
            },
            'conv_block_3':{
                'filters': 64,
                'kernel_size': 3,
                'strides': 1,
                'padding': 'same',
                'activation': 'relu',
            },
            'pool_block_3': {
                "pool_size": 2
            },
            'flatten_layer':{

            },
            "flatten_dropout":{
                'rate':0.5
            }
    }


    def make_inputs(self,seq_length) -> list:
        input_left = keras.layers.Input(
            name='text_left',
            shape=(seq_length,)
        )
        input_right = keras.layers.Input(
            name='text_right',
            shape=(seq_length,)
        )
        return [input_left, input_right]

    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size
        config = self.hyper_parameters
        embedding_model = self.embedding.embed_model

        conv1d_layer = L.Conv1D(**config["conv_1d_layer"])
        matching_layer = MatchingLayer(**config['matching_layer'])
        flatten_layer = L.Flatten(**config["flatten_layer"])
        flatten_dropout = L.Dropout(**config['flatten_dropout'])




        # Left input and right input.
        input_left,input_right = self.make_inputs(self.sequence_length)
        input_left_embed ,input_right_embed = embedding_model(input_left),embedding_model(input_right)

        input_left_conv1d = conv1d_layer(input_left_embed)
        input_right_conv1d = conv1d_layer(input_right_embed)

        embed_cross = matching_layer([input_left_conv1d,input_right_conv1d])

        for i in range(1,int(config['num_blocks']) +1):
            conv2d_layer = L.Conv2D(**config['conv_block_'+ str(i)])
            pool2d_layer = L.MaxPooling2D(**config["pool_block_"+str(i)])
            embed_cross = conv2d_layer(embed_cross)
            embed_cross = pool2d_layer(embed_cross)

        embed_flat = flatten_layer(embed_cross)
        embed_flat = flatten_dropout(embed_flat)


        output = self.make_output_layer(output_dim)(embed_flat)


        self.tf_model = keras.Model([input_left,input_right],output)









import codecs
import os
import pathlib
import re

from setuptools import find_packages, setup




__name__ = 'nlp_tools'
__author__ = "fanfanfeng"
__copyright__ = "Copyright 2020, fanfanfeng"
__credits__ = []
__license__ = "Apache License 2.0"
__maintainer__ = "fanfanfeng"
__email__ = "544855237@qq.com"
__url__ = 'https://github.com/BrikerMan/Kashgari'
__description__ = '参考Kashgari项目，然后会扩充一些自己做的项目'

__version__ = '1.0.1'
README = __description__

#with codecs.open('requirements.txt', 'r', 'utf8') as reader:
#    install_requires = list(map(lambda x: x.strip(), reader.readlines()))
install_requires = []

# setup(
#     name=__name__,
#     version=__version__,
#     description=__description__,
#     python_requires='>3.6',
#     long_description=README,
#     long_description_content_type="text/markdown",
#     author=__author__,
#     author_email=__email__,
#     url=__url__,
#     packages=find_packages(exclude=('tests',)),
#     install_requires=install_requires,
#     include_package_data=True,
#     zip_safe=True
#     #license=LGPL,
#
# )

setup(
    name="nlp_tools",
    version='0.2.2',
    description='工程提取字段相关的一些定义以及一些常用函数封装',
    author='qiufengfeng',
    install_requires = [],
    author_email='544855237@qq.com',
    package_dir={'nlp_tools': 'nlp_tools'},
    packages=find_packages(),
    include_package_data=True,
    #package_data={'infoprocessor': ['data/*']},
    license='LGPL',
    zip_safe=False
)





#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   confusion_seaborn.py    
@Contact :   544855237@qq.com

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/12/20 下午3:56   qiufengfeng      1.0         None
'''

# import lib
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix

sns.set_style({"font.sans-serif":['SimHei','Arial']})


def plot_confusion_mat(gold_lables,pred_labels,labels,title="混淆矩阵",normalize=None):
    """
    @param gold_lables:真实样本
    @param pred_labels:预测样本
    @param labels:预测标签样本
    @param title:生成的图片标题
    @return:
    """
    plt.figure(figsize=(len(labels),len(labels)))
    _confmat = confusion_matrix(gold_lables,pred_labels,normalize=normalize)
    print(_confmat)
    if normalize is not None:
        _heatmap = sns.heatmap(_confmat,vmax=1,vmin=0,square=True,annot=True,cmap='Blues',xticklabels=labels,yticklabels=labels,fmt='g')
    else:
        _heatmap = sns.heatmap(_confmat,square=True,annot=True,cmap='Blues',xticklabels=labels,yticklabels=labels,fmt='g')
    _heatmap.tick_params(which='major', labelbottom=False, labeltop=True)
    figure = _heatmap.get_figure()
    figure.show()
    return figure




from typing import List,Tuple
from nlp_tools.metrics.sequence_labeling import end_of_chunk,start_of_chunk

def get_entities(seq: List[str], *, suffix: bool = False) -> List[Tuple[str, int, int]]:
    """Gets entities from sequence.
        Args:
            seq: sequence of labels.
            suffix:
        Returns:
            list: list of (chunk_type, chunk_start, chunk_end).
        Example:
            >>> from nlp_tools.utils.ner_utils import get_entities
            >>> seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
            >>> get_entities(seq)
            [('PER', 0, 1), ('LOC', 3, 3)]
        """
    prev_tag = 'O'
    prev_type = ''
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq + ['O']):
        if suffix:
            tag = chunk[-1]
            type_ = chunk.split('-')[0]
        else:
            tag = chunk[0]
            type_ = chunk.split('-')[-1]

        if end_of_chunk(prev_tag, tag, prev_type, type_):
            chunks.append((prev_type, begin_offset, i -1))
        if start_of_chunk(prev_tag, tag, prev_type, type_):
            begin_offset = i
        prev_tag = tag
        prev_type = type_
    return chunks

def format_ner_result(pred_labels,tokened_sentences,orgin_sentences):
    new_res = [get_entities(seq) for seq in pred_labels]
    final_res = []
    for index, seq in enumerate(new_res):
        seq_data = []
        for entity in seq:
            res_entities: List[str] = []
            for i, e in enumerate(tokened_sentences[index][entity[1]:entity[2] + 1]):
                # Handle bert tokenizer
                if e.startswith('##') and len(res_entities) > 0:
                    res_entities[-1] += e.replace('##', '')
                else:
                    res_entities.append(e)


            value = "".join(res_entities)

            seq_data.append({
                "entity": entity[0],
                "start": entity[1],
                "end": entity[2],
                "value": value,
            })

        final_res.append({
            'sentence': orgin_sentences[index],
            'labels': seq_data
        })
    return final_res

def output_ner_results(origin_texts,predict_labels):
    '''

    '''
    result =[]
    for origin_text,predict_label in zip(origin_texts,predict_labels):
        entity_dict = {}
        entity_dict['sentence'] = origin_text
        labels = []
        for (start_id,end_id,label) in predict_label:

            labels.append({
                "entity": label,
                "start": start_id,
                "end": end_id,
                "value": origin_text[start_id:end_id+1],
            })
        entity_dict['labels'] = labels
        result.append(entity_dict)

    return result


def span_extract_item(start_logits, end_pred):
    '''return type list
    [(labeltype，start,end)]
    '''
    S = []
    for i, s_l in enumerate(start_logits):
        if s_l == 0:
            continue
        for j, e_l in enumerate(end_pred[i:]):
            if s_l == e_l:
                S.append((s_l, i, i + j))
                break
    return S






from typing import Dict,Any
from tensorflow import keras
from nlp_tools.layers import L
from nlp_tools.tasks.classification.abc_model import ABCClassificationModel
class BiLSTM_Model(ABCClassificationModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'layer_bi_lstm': {
                'units': 128,
                'return_sequences': False
            },
            'layer_output': {

            }
        }

    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        # build model structure in sequent way
        layer_stack = [
            L.Bidirectional(L.LSTM(**config['layer_bi_lstm'])),
            L.Dense(output_dim, **config['layer_output']),
            self._activation_layer()
        ]

        tensor = embed_model.output
        for layer in layer_stack:
            tensor = layer(tensor)

        self.tf_model: keras.Model = keras.Model(embed_model.inputs, tensor)



import os
import json
from abc import ABC
from typing import List,Dict,Any,Union
from sklearn import metrics as sklearn_metrics

import nlp_tools
from nlp_tools.layers import L


from nlp_tools.tasks.abs_task_model import ABCTaskModel
from nlp_tools.types import TextSamplesVar,ClassificationLabelVar,MultiLabelClassificationLabelVar

from nlp_tools.embeddings import ABCEmbedding,TransformerEmbedding,BertEmbedding
from nlp_tools.optimizer import MultiOptimizer
from tensorflow.keras.optimizers import Adam

#from transformers.optimization_tf import AdamWeightDecay
from nlp_tools.loss.r_drop_loss import RDropLoss

from nlp_tools.metrics.plot.confusion_seaborn import plot_confusion_mat

class ABCClassificationModel(ABCTaskModel,ABC):
    """
    Abstract Classification Model
    """

    __task__ = 'classification'

    def to_dict(self) -> Dict:
        info = super(ABCClassificationModel, self).to_dict()
        return info

    def __init__(self,
                 text_processor,
                 label_processor,
                 embedding: ABCEmbedding = None,
                 max_sequence_length: int = None,
                 hyper_parameters: Dict[str, Dict[str, Any]] = None,
                 train_sequece_length_as_max_sequence_length=False,
                 use_FGM=True,
                 use_rdrop=True):

        super(ABCClassificationModel, self).__init__(text_processor=text_processor,label_processor=label_processor,
                                                     embedding=embedding,max_sequence_length=max_sequence_length,hyper_parameters=hyper_parameters,
                                                     train_sequece_length_as_max_sequence_length=train_sequece_length_as_max_sequence_length,
                                                     use_FGM=use_FGM,use_rdrop=use_rdrop)
    def _activation_layer(self) -> L.Layer:
        # if self.multi_label:
        #     return L.Activation('sigmoid')
        # else:
        return L.Activation('softmax')




    def build_model_arc(self) -> None:
        raise NotImplementedError

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:

        if loss is None:
            from nlp_tools.loss import multi_category_focal_loss2_fixed
            from tensorflow.keras.losses import SparseCategoricalCrossentropy
            from nlp_tools.loss import categorical_focal_loss
            loss = SparseCategoricalCrossentropy()
            #loss = multi_category_focal_loss2_fixed

        if type(self.embedding) in [TransformerEmbedding, BertEmbedding]:
            total_layers = self.tf_model.layers
            transfomer_layers = self.embedding.embed_model.layers
            no_transformer_layers = [layer for layer in total_layers if layer not in transfomer_layers]
            optimizer_list = [
                Adam(),
                Adam(learning_rate=1e-5)
            ]
            optimizers_and_layers = [(optimizer_list[0], no_transformer_layers), (optimizer_list[1], transfomer_layers)]
            optimizer = MultiOptimizer(optimizers_and_layers)
        else:
            optimizer = Adam(1e-5)
        if self.use_rdrop:
            if type(loss) == list:
                loss = [RDropLoss(i) for i in loss]
            else:
                loss = RDropLoss(loss)

        #optimizer = Adam()#AdamW(weight_decay=0.0)
        if metrics is None:
            metrics = ['accuracy']

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)




    def predict(self,
                x_data: TextSamplesVar,
                *,
                batch_size: int = 32,
                truncating: bool = False,
                multi_label_threshold: float = 0.5,
                predict_kwargs: Dict = None,
                return_labels=True) -> Union[ClassificationLabelVar, MultiLabelClassificationLabelVar]:
        if predict_kwargs is None:
            predict_kwargs = {}

        with nlp_tools.utils.custom_object_scope():
            if truncating:
                seq_length = self.max_sequence_length
            else:
                seq_length = None
            tensor = self.text_processor.transform(x_data,seq_length=seq_length)
            pred = self.tf_model.predict(tensor, batch_size=batch_size, **predict_kwargs)
            pred_argmax = pred.argmax(-1)
            if return_labels:
                res = self.label_processor.inverse_transform(pred_argmax)
            else:
                res = pred_argmax

        return res

    def evaluate(self,
                 valid_data,
                 *,
                 batch_size: int = 32,
                 digits: int = 4,
                 multi_label_threshold: float = 0.5,
                 result_save_path = None):
        x_data = [item[0] for item in valid_data]
        y_data = [item[1] for item in valid_data]
        y_pred_ids = self.predict(x_data,
                              batch_size=batch_size,
                              multi_label_threshold=multi_label_threshold,
                              return_labels=False)

        valid_y_ids = self.label_processor.transform([y for y in y_data])
        label_names = [key for key in self.label_processor.vocab2idx.keys()]

        original_report = sklearn_metrics.classification_report(valid_y_ids,
                                                                y_pred_ids,
                                                                digits=digits,
                                                                target_names=label_names)
        print(original_report)
        if result_save_path:
            if not os.path.exists(result_save_path):
                os.mkdir(result_save_path)
            confusion_save_path = os.path.join(result_save_path, 'confusion.jpg')
            classify_report_path = os.path.join(result_save_path, 'classify_report.txt')
            figure = plot_confusion_mat(valid_y_ids, y_pred_ids, label_names)
            figure.savefig(confusion_save_path, dpi=600)

            with open(classify_report_path,'w',encoding='utf-8') as fwrite:
                if type(original_report) == dict:
                    original_report = json.dumps(original_report,indent=4,ensure_ascii=False)
                fwrite.write(original_report)




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   autoembedding.py
@Contact :   544855237@qq.com

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/12/22 下午4:36   qiufengfeng      1.0         None
'''

# import lib
from transformers import TFAutoModel,AutoConfig,AutoTokenizer
import tensorflow as tf
from tensorflow.keras import layers,Model
import numpy as np

from nlp_tools.embeddings.abc_embedding import ABCEmbedding
from nlp_tools.embeddings.hugginface.prefix_prompt.configuration_prefix_prompt_roberta import PrefixPromptConfig
from typing import Dict,Any


class AutoEmbedding(ABCEmbedding):
    def to_dict(self) -> Dict[str, Any]:
        info_dic = super(AutoEmbedding, self).to_dict()
        return info_dic

    def __init__(self,pretrained_model_path,inputs_keys,**kwargs):
        super(AutoEmbedding, self).__init__(**kwargs)

        self.pretrained_model_path = pretrained_model_path
        self.inputs_keys = inputs_keys
        self.max_position = 512

    def make_inputs_nodes(self):
        input_dict = {}
        for key in self.inputs_keys:
            input_dict[key] = layers.Input(shape=(None,),dtype=tf.int32,name=key)
        return input_dict






    def build_embedding_model(self) -> None:
        inputs = self.make_inputs_nodes()
        input_list = [value for key,value in inputs.items()]


        autoBertModel = TFAutoModel.from_pretrained(self.pretrained_model_path)
        output = autoBertModel.bert(inputs,training=True)

        prefixEncoderModel = Model(input_list,output.last_hidden_state)


        self.embed_model = prefixEncoderModel
        self.embedding_size = prefixEncoderModel.output.shape[-1]




if __name__ == '__main__':

    model_name = "bert-base-uncased"
    from transformers import  AutoTokenizer
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    a = AutoEmbedding(model_name,tokenizer.model_input_names)
    a.build_embedding_model()




from nlp_tools.processors.class_processor import ClassificationProcessor
from nlp_tools.processors.sequence_processor import SequenceProcessor
from nlp_tools.processors.dialogue_sequence_processor import DialogueSequenceProcessor



# encoding: utf-8


import collections
import operator
from typing import Dict, List, Any, Optional, Union

import numpy as np
from tensorflow.keras.preprocessing.sequence import pad_sequences

from nlp_tools.generators import CorpusGenerator
from nlp_tools.logger import logger
from nlp_tools.processors.abc_sentence_processor import ABCSentenceProcessor
from nlp_tools.types import TextSamplesVar
import tqdm
import re


class HierarchicalSequenceProcessor(ABCSentenceProcessor):
    """
    hierrarchical_attention_networks处理相关类，需要吧一段文本根据句号，逗号等分割城多个短句
    """



    def __init__(self,
                 **kwargs: Any) -> None:
        """

        Args:
            vocab_dict_type: initial vocab dict type, one of `text` `labeling`.
            **kwargs:
        """
        super(HierarchicalSequenceProcessor, self).__init__(**kwargs)



    def transform(self,samples: TextSamplesVar,seq_length: int = None,doc_length:int = None) :

        split_char = "，|。|！|？"
        split_samples = []

        for sample in samples:
            split_list = re.split(split_char,sample)
            split_list = [i for i in split_list if i.strip()!=""]

            split_samples.append(split_list)


        if seq_length is None and self.max_sentence_length is not None:
            seq_length = self.max_sentence_length
        #seq_length = 50

        if seq_length is None:
            seq_length = np.max([max([len(self.text_tokenizer.tokenize(i)) for i  in j]) for j in split_samples])

        if self.embedding_max_position is not None and self.embedding_max_position < seq_length:
            seq_length = self.embedding_max_position



        doc_length = 20


        tokenids = []
        for doc_seq in split_samples:
            item = []
            sub_doc_masking = []
            for seq in doc_seq:
                seq_tokenids,seq_segment_ids = self.text_tokenizer.encode(seq,maxlen=seq_length)
                if len(seq_tokenids) > seq_length:
                    seq_tokenids = seq_tokenids[:seq_length]
                else:
                    seq_tokenids += [0] * (seq_length - len(seq_tokenids))
                #seq_tokenids = pad_sequences(seq_tokenids, seq_length, padding='post', truncating='post')
                item.append(seq_tokenids)
                sub_doc_masking.append(1)



            if len(item) < doc_length:
                for _ in range(doc_length - len(item)):
                    item.append([0 for _ in range(seq_length)])
                    sub_doc_masking.append(0)
            tokenids.append(item)





        token_ids = np.array(tokenids)
        token_ids = np.reshape(token_ids, (-1, doc_length , seq_length))

        #segment_ids = np.zeros_like(token_ids)
        return token_ids


    def inverse_transform(self,
                          labels: Union[List[List[int]], np.ndarray],
                          *,
                          lengths: List[int] = None,
                          threshold: float = 0.5,
                          **kwargs: Any) -> List[List[str]]:
        raise NotImplementedError("暂时没有实现！！")


    def get_max_seq_length_from_corpus(self,
                                   generators: List[CorpusGenerator],
                                   cover_rate: float = 0.95) -> int:
        """
        Calculate proper sequence length according to the corpus

        Args:
            generators:
            use_label:
            cover_rate:

        Returns:

        """
        seq_lens = []
        for gen in generators:
            for (sentence, _) in tqdm.tqdm(gen.sample(), desc="Calculating sequence length"):
                seq_lens.append(len(self.text_tokenizer.tokenize(sentence)))
        if cover_rate == 1.0:
            target_index = -1
        else:
            target_index = int(cover_rate * len(seq_lens))
        sequence_length = sorted(seq_lens)[target_index]
        logger.debug(f'Calculated sequence length = {sequence_length}')
        return sequence_length

    def update_length_info(self,embedding_max_position=None,max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length





if __name__ == "__main__":
    pass



import json
import os
import pathlib
from abc import ABC, abstractmethod
from typing import Dict, Any, TYPE_CHECKING, Union
import functools

import tensorflow as tf

import nlp_tools
from nlp_tools.embeddings import ABCEmbedding
from nlp_tools.logger import logger
from nlp_tools.utils import load_data_object
from nlp_tools.utils.training_tips import creat_FGM
from nlp_tools.generators import CorpusGenerator, BatchGenerator



class ABCTaskModel(ABC):

    def __init__(self,
                 text_processor,
                 label_processor,
                 embedding: ABCEmbedding = None,
                 max_sequence_length: int = None,
                 hyper_parameters: Dict[str, Dict[str, Any]] = None,
                 train_sequece_length_as_max_sequence_length = False,
                 use_FGM = True,
                 use_rdrop=True):
        """

        Args:
            embedding: embedding object
            max_sequence_length: target sequence length
            hyper_parameters: hyper_parameters to overwrite
        """
        super(ABCTaskModel, self).__init__()
        if hyper_parameters is None:
            hyper_parameters = self.default_hyper_parameters()

        self.tf_model = None
        self.embedding = embedding
        self.hyper_parameters = hyper_parameters


        self.text_processor = text_processor
        self.label_processor = label_processor

        self.max_sequence_length = max_sequence_length
        self.train_max_sequence_length = max_sequence_length
        self.train_sequece_length_as_max_sequence_length = train_sequece_length_as_max_sequence_length
        self.use_FGM = use_FGM
        self.use_rdrop = use_rdrop

        ## 如果分词器支持keep_tokens，且embedding也支持keep_tokens，则需要吧分词器的keep_tokens跟新到
        if self.text_processor:
            if hasattr(self.text_processor.text_tokenizer,"keep_tokens") and self.text_processor.text_tokenizer.keep_tokens != "" and hasattr(self.embedding,"keep_tokens"):
                self.embedding.keep_tokens = self.text_processor.text_tokenizer.keep_tokens


        self.embedding_max_position = None
        if self.embedding and  hasattr(self.embedding,"max_position"):
            self.embedding_max_position = self.embedding.max_position

        self.model_node_names = {}






    def to_dict(self) -> Dict[str, Any]:
        #model_json_str = self.tf_model.to_json()

        save_json = {
            'tf_version': tf.__version__,
            'nlp_tools_version':nlp_tools.__version__,
            '__class_name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
            'config': {
                'hyper_parameters': self.hyper_parameters,
                'max_sequence_length': self.max_sequence_length,
                'use_FGM':self.use_FGM,
                'use_rdrop':self.use_rdrop
            },
            'label_processor': self.label_processor.to_dict(),
            #'tf_model': json.loads(model_json_str),
            'model_node_names':self.model_node_names
        }

        if self.embedding:
            save_json['embedding'] = self.embedding.to_dict()

        if self.text_processor:
            save_json['text_processor'] = self.text_processor.to_dict()

        return save_json

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        raise NotImplementedError

    def save(self, model_path: str) -> str:
        pathlib.Path(model_path).mkdir(exist_ok=True, parents=True)
        model_path = os.path.abspath(model_path)

        # 将inputs和outputs 节点的name保存下来
        self.model_node_names = self.get_input_and_output_names_from_model()

        if(hasattr(self.text_processor.text_tokenizer,"change_model_path")):
            self.text_processor.text_tokenizer.change_model_path(os.path.join(model_path,'text_tokenizer'))

        with open(os.path.join(model_path,'model_config.json'), 'w',encoding='utf-8') as f:
            f.write(json.dumps(self.to_dict(),indent=2,ensure_ascii=False))
            f.close()

        tf.keras.models.save_model(self.tf_model,os.path.join(model_path,'model_weights'),include_optimizer=False)

        logger.info('model saved to {}'.format(os.path.abspath(model_path)))
        return model_path

    @classmethod
    def load_model(cls,model_path:str,**kwargs):
        model_config_path = os.path.join(model_path,'model_config.json')
        model_config = json.loads(open(model_config_path,'r',encoding='utf-8').read())


        if 'model_path' in model_config['text_processor']['text_tokenizer']['config']:
            model_config['text_processor']['text_tokenizer']['config']['model_path'] = os.path.join(model_path,'text_tokenizer')

        model_config['config']['text_processor'] = load_data_object(model_config['text_processor'])
        model_config['config']['label_processor'] = load_data_object(model_config['label_processor'])
        #model_config['config']['embedding'] = load_data_object(model_config['embedding'])
        model = load_data_object(model_config)
        model.tf_model = tf.keras.models.load_model(os.path.join(model_path,'model_weights'),custom_objects=nlp_tools.custom_objects)
        #model.tf_model = tf.keras.models.load_model(os.path.join(model_path, 'model_weights'))
        return model


    def build_model(self,
                    x_data: Any,
                    y_data: Any) -> None:
        train_gen = CorpusGenerator(x_data, y_data)
        self.build_model_generator([train_gen])



    def build_model_arc(self) -> None:
        raise NotImplementedError

    def compile_model(self) -> None:
        raise NotImplementedError

    def build_model_generator(self,generators) -> None:

        self.label_processor.build_vocab_generator(generators)
        if self.train_max_sequence_length is None and self.text_processor:
            self.train_max_sequence_length = self.text_processor.get_max_seq_length_from_corpus(generators,cover_rate=1.0)

        if self.train_sequece_length_as_max_sequence_length:
            self.max_sequence_length = self.train_sequece_length_as_max_sequence_length

        self.label_processor.update_length_info(embedding_max_position=self.embedding_max_position,max_sentence_length=self.max_sequence_length)

        if self.text_processor:
            self.text_processor.update_length_info(embedding_max_position=self.embedding_max_position,max_sentence_length=self.max_sequence_length)

        if self.embedding:
            self.embedding.build_embedding_model()


        if self.tf_model is None:
            self.build_model_arc()
            self.compile_model()

            # 替换model.train_step 方法即可,并且删除原有的 train_function方法
            if self.use_FGM:
                train_step = creat_FGM()
                self.tf_model.train_step = functools.partial(train_step, self.tf_model)
                self.tf_model.train_function = None


    def get_input_and_output_names_from_model(self):
        '''从模型中获取到输入和输出的node 的名称'''
        input_names = [node.name for node in self.tf_model.inputs]
        output_names = [node.name for node in self.tf_model.outputs]
        node_names_dict = {}
        node_names_dict['inputs'] = input_names
        node_names_dict['outputs'] = output_names
        return node_names_dict


    def fit(self,
            train_data,
            validate_data=None,
            batch_size: int = 64,
            epochs: int = 5,
            callbacks = None,
            fit_kwargs: Dict = None,
            generator=None) -> 'tf.keras.callbacks.History':

        if fit_kwargs is None:
            fit_kwargs = {}

        if generator == None:
            generator = BatchGenerator

        train_generator = generator(train_data,
                                   text_processor=self.text_processor,
                                   label_processor=self.label_processor,
                                   seq_length=self.train_max_sequence_length,
                                   batch_size=batch_size,
                                    use_rdrop=self.use_rdrop)
        if validate_data:
            valid_generator = generator(validate_data,
                                     text_processor=self.text_processor,
                                     label_processor=self.label_processor,
                                     seq_length=self.train_max_sequence_length,
                                     batch_size=batch_size,
                                     use_rdrop=self.use_rdrop)
            fit_kwargs['validation_data'] = valid_generator.forfit()
            fit_kwargs['validation_steps'] = len(valid_generator)
        else:
            valid_generator = None


        return self.fit_generator(train_sample_gen=train_generator,
                                  valid_sample_gen=valid_generator,
                                  epochs=epochs,
                                  callbacks=callbacks,
                                  fit_kwargs=fit_kwargs)

    def fit_generator(self,
                      train_sample_gen,
                      valid_sample_gen,
                      epochs: int = 5,
                      callbacks = None,
                      fit_kwargs: Dict = None) -> 'tf.keras.callbacks.History':
        self.build_model_generator([g for g in [train_sample_gen, valid_sample_gen] if g])
        return self.tf_model.fit(train_sample_gen.forfit(),
                                 steps_per_epoch=len(train_sample_gen),
                                 epochs=epochs,
                                 callbacks=callbacks,
                                 **fit_kwargs)


import numpy as np
import tensorflow as tf

from nlp_tools.embeddings.abc_embedding import ABCEmbedding
from nlp_tools.layers import L

class GRUEncoder(tf.keras.Model):
    def __init__(self, embedding: ABCEmbedding, hidden_size: int = 1024):
        super(GRUEncoder, self).__init__()
        self.embedding = embedding
        self.hidden_size = hidden_size
        self.gru = tf.keras.layers.GRU(hidden_size,
                                       return_sequences=True,
                                       return_state=True,
                                       recurrent_initializer='glorot_uniform')

    def call(self, x: np.ndarray, hidden: np.ndarray) -> np.ndarray:
        if self.embedding.segment:
            x = (x, tf.zeros(x.shape))
        x = self.embedding.embed_model(x)
        output, state = self.gru(x, initial_state=hidden)
        return output, state

    def model(self) -> tf.keras.Model:
        x1 = L.Input(shape=(None,))
        x2 = L.Input(shape=(self.hidden_size,))
        return tf.keras.Model(inputs=[x1, x2],
                              outputs=self.call(x1, x2),
                              name='GRUEncoder')


if __name__ == "__main__":
    pass
